%% BFGS method
clear;
close all;
clc;

plotFigures = false;
% plotFigures = true;

% Number of mass points
N     = 40;
tol   = 1E-3;
maxit = 10000;

% Chain end points
param.xi = [-2 1];  % initial (y1, z1)
param.xf = [ 2 1];  % final   (yN, zN)

% Definining constants in a structure
param.L = 1;
param.D = (70/40)*N;
param.m = 4/N;
param.g = 9.81;
param.N = N;

% Initial guess: linear interpolation between the tips
y = linspace(param.xi(1), param.xf(1), N+2)';
z = linspace(param.xi(2), param.xf(2), N+2)';

x = reshape([y(2:end-1) z(2:end-1)].',2*N,1); % Eliminate the tips from the decision variables and stack them together

% Hessian approximation
% Bk = 10000*eye(length(x)); % this works also without line search
Bk = eye(length(x));

% Get the objective value and the gradient at initial point
[F, J] = finite_difference(@hc_obj, x, param);
% A check whether the function you wrote is  correct
if F > 46.93 || F < 46.92 || norm(J) > 6.21 || norm(J) < 6.20  
    error('The outputs from your function finite_difference.m for the given initial point are not correct.')
end

% Printing header: iterate number, gradient norm, objective value, step norm, stepsize
fprintf('It.  \t | ||grad_f||| f\t\t | ||dvar||\t | t  \n');

gradnorm_gradient = norm(J);
tic
% Iterative loop
for k = 1 : maxit
    % Obtaining search direction
    dx = -Bk \ J';
    
    % We check Armijo's condition to find a step length
    t     = 1.0;    % Initial step length
    beta  = 0.8;    % Shrinking factor
    gamma = 0.1;    % Minimal decrease requirement
    
    x_new = x + t * dx; % Candidate for the next step

    % Iterate on Armijo's condition
    while hc_obj(x_new, param) > F + gamma * t * J * dx
        t = beta * t;
        x_new = x + t * dx;
    end
    
    % Assign the step
    [F_new, J_new] = finite_difference(@hc_obj, x_new, param);
    
    % Updating variables
    x = x_new;
    J = J_new;
    F = F_new;

    % Every 10 iterations print the header again
    if mod(k,10) == 0
        fprintf('\n');
        fprintf('It.  \t | ||grad_f||| f\t\t | ||dvar||\t | t  \n');
    end
    
    % Print some useful information
    fprintf('%d\t | %f\t | %f\t | %f\t | %f \n', k, norm(J), F, norm(dx), t);
    
    % plotting
    if plotFigures
        y = x(1:2:2*N);
        z = x(2:2:2*N);
        % Plotting 
        figure(1)
        subplot(2,1,1), plot(y, z,'b--');hold on;
        subplot(2,1,1), plot(y, z, 'Or'); hold off;
        xlim([-2, 2])
        ylim([ -3, 1])
        title('Position of chain at current iterate')
        
        subplot(2,1,2), plot(dx)
        title('full step of each optimization variable (dz_i)')
        drawnow;
    end
    if norm(J) < tol
        disp('Convergence achieved.');
        break
    end
%     keyboard
    %pause(0.05)
end
toc
if plotFigures
    disp('(set ''plotFigures = false'' for accurate times)')
end
% Plotting optimal solution 
figure(1)
y = x(1:2:2*N);
z = x(2:2:2*N);
subplot(2,1,1), plot(y, z, 'b--');hold on;
title('Position of chain at optimal solution')
subplot(2,1,1), plot(y, z, 'Or'); hold off;
xlim([-2, 2])
ylim([ -3, 1])